"""
Implements a camunda client.
"""
import logging
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urljoin

from django.conf import settings
from django.utils.module_loading import import_string

import inflection
import requests

from .models import CamundaConfig

logger = logging.getLogger(__name__)


def get_client_class() -> type:
    client_class = getattr(settings, "CAMUNDA_CLIENT_CLASS", "django_camunda.client.Camunda")
    return import_string(client_class)


class Camunda:
    def __init__(self, config: Optional[CamundaConfig] = None):
        config = config or CamundaConfig.get_solo()
        self.root_url = config.api_root

    def request(self, path: str, method="GET", *args, **kwargs):
        assert not path.startswith("/"), "Provide relative API paths"
        url = urljoin(self.root_url, path)

        # add the API headers, so that Camunda can use the tokens. Essentially
        # we're forwarding Auth
        headers = kwargs.pop("headers", {})
        headers.update(self.get_extra_headers(headers))
        kwargs["headers"] = headers

        json = kwargs.get("json")
        if json:
            self.preprocess_json(json)

        _ref = self.before_request(method, url, *args, **kwargs)

        response = requests.request(method, url, *args, **kwargs)
        response_data = None

        try:
            response.raise_for_status()
            if response.content:
                response_data = response.json()

                if isinstance(response_data, (dict, list)):
                    self.postprocess_response_data(response_data)
            return underscoreize(response_data)
        except Exception:
            try:
                # see if we can grab any extra output
                response_data = response.json()
            except Exception:
                pass
            logger.exception("Error: %r", response_data)
            raise
        finally:
            self.after_request(_ref, response, response_data)

    # HOOKS for subclasses

    def before_request(self, method: str, url: str, *args, **kwargs) -> Any:
        pass

    def after_request(self, ref: Any, response, response_data) -> None:
        pass

    def get_extra_headers(self, headers: dict) -> dict:
        return {}

    def preprocess_json(self, json: dict) -> None:
        pass

    def postprocess_response_data(self, data: Union[list, dict]) -> None:
        pass


def underscoreize(data: Union[List, Dict, str, None]) -> Union[List, Dict, str, None]:
    if isinstance(data, list):
        return [underscoreize(item) for item in data]

    if isinstance(data, dict):
        new_data = {}
        for key, value in data.items():
            new_key = inflection.underscore(key)
            new_data[new_key] = underscoreize(value)
        return new_data

    return data



import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional, Union

from dateutil.parser import parse

CONVERTERS = {
    type(None): lambda x: None,
    uuid.UUID: lambda value: uuid.UUID(value),
    datetime: parse,
}


class Model:
    def __post_init__(self):
        self._type_cast()

    def _type_cast(self):
        for attr, typehint in self.__annotations__.items():
            value = getattr(self, attr)

            if typehint is None:
                typehint = type(None)

            # support for Optional
            if hasattr(typehint, "__origin__") and typehint.__origin__ is Union:
                typehint = typehint.__args__

                if value is None:
                    continue

                # Optional is ONE type combined with None
                typehint = next(t for t in typehint.__args__ if t is not None)

            if isinstance(value, typehint):
                continue

            converter = CONVERTERS[typehint]
            setattr(self, attr, converter(value))


@dataclass
class Task(Model):
    id: uuid.UUID
    name: str
    assignee: None
    created: datetime
    due: Optional[datetime]
    follow_up: None
    delegation_state: None
    description: None
    execution_id: str
    owner: None
    parent_task_id: None
    priority: int
    process_definition_id: str
    process_instance_id: uuid.UUID
    task_definition_key: str
    case_execution_id: None
    case_instance_id: None
    case_definition_id: None
    suspended: bool
    form_key: None
    tenant_id: None

    def claim(self) -> None:
        from bing.service.camunda import claim_task

        claim_task(self.id)

    def complete(self, variables: Dict[str, Any]) -> None:
        from bing.service.camunda import complete_task

        complete_task(self.id, variables)
